/*
 * ftp_get_file
 *
 * This plugin attempts to retrieve /etc/passwd, and warns the user 
 * if the accounts of this file have encrypted passwd, and if 
 * there are others accounts than the system ones...
 *
 * This plugin was written by Renaud Deraison and is 
 * distributed under the GPL
 */
#include <includes.h>

#define NAME "ftp get /etc/passwd"
#define DESCRIPTION "\
Some badly configured ftp servers are displaying the\n\
actual /etc/passwd file to the world. This gives away\n\
sensitive informations such as the user names, and even\n\
sometimes their encrypted passwd\n\n\
Risk factor : medium/high"
#define COPYRIGHT "this plugin is distributed under the GPL"
#define SUMMARY "reads /etc/passwd via ftp"


int ftp_log_in(int soc, char * username, char * passwd);
int get_pasv_address(int soc, struct sockaddr_in * addr);

PlugExport int plugin_init(struct arglist *desc);
PlugExport int plugin_init(struct arglist *desc)
{
        plug_set_name(desc, NAME);
        plug_set_description(desc, DESCRIPTION);
        plug_set_summary(desc, SUMMARY);
        plug_set_copyright(desc, COPYRIGHT);
        plug_set_category(desc, ACT_ATTACK);
        plug_set_family(desc, "FTP");
        return(0);
}

PlugExport int plugin_run(struct arglist * args);
PlugExport int plugin_run(struct arglist * args)
{
 int soc[2];
 char * buffer;
 char *t=NULL, *t2;
 fd_set read_set;
 struct timeval tv = {2,0};
 struct sockaddr_in addr;
 int finished = 0;
 char * report;
 int a = 1;
 if(host_get_port_state(args, 21)<=0)return(0);
 if((soc[0] = open_sock_tcp(args, 21))<0)return(0);
 if(ftp_log_in(soc[0], "anonymous", "joe@"))return(0);
 if(get_pasv_address(soc[0], &addr))return(0);
 send(soc[0], "TYPE I\n", 7, 0);
 buffer = emalloc(255);
 recv_line(soc[0], buffer, 255);
 send(soc[0], "RETR /etc/passwd\n", 17, 0);
 soc[1] = socket(AF_INET, SOCK_STREAM, 0);
 if(connect(soc[1], (struct sockaddr *)&addr, sizeof(struct sockaddr_in)))
 return(0);
 bzero(buffer, 255);

 recv_line(soc[0], buffer, 255);
 if(strncmp(buffer, "150", 3))return(0);

 while(!finished)
 {
 if(a)
  {
   bzero(buffer, 255);
   recv(soc[1], buffer, 255, 0);
   if(t){
     t2 = emalloc(strlen(t)+strlen(buffer)+1);
     sprintf(t2, "%s%s", t, buffer);
     efree(&t);
     t = t2;
     }
   else
     {
     t2 = emalloc(strlen(buffer)+1);
     sprintf(t2, "%s", buffer);
     t = t2;
     }
   finished = 1;
   ioctl(soc[1], FIONREAD, &a);
  }
 else finished = 1;
 
 }
 
  
 if(dangerous_password_file(t))
 {
  char * report = emalloc(255+strlen(t));
  sprintf(report, "\
The file /etc/passwd obtained via anonymous FTP\n\
seems to give away too much informations. You should\n\
change it as soon as possible :\n%s\n", t);
  post_hole(args, 21, report);
  efree(&report);
 }
 if(t)efree(&t);
 return(0);
}
 


int ftp_log_in(int soc, char * username, char * passwd)
{
 char * buf;
 fd_set read_set;
 struct timeval tv = {10,0};
 
 FD_ZERO(&read_set);
 FD_SET(soc, &read_set);
 select(soc+1, &read_set, NULL,NULL, &tv);
 if(!FD_ISSET(soc, &read_set))return(1);
 
 buf = emalloc(512);
 recv_line(soc, buf, 512);
 if(strncmp(buf, "220", 3))
 { 
  efree(&buf);
  return(1);
 }
 
 while(buf[3]=='-')recv_line(soc, buf, 512);
 bzero(buf, 512);
 sprintf(buf, "USER %s\n", username);
 send(soc, buf, strlen(buf), 0);
 recv_line(soc, buf, 512);
 if(strncmp(buf, "331", 3))
 {
  efree(&buf);
  return(1);
 }
 bzero(buf, 512);
 sprintf(buf, "PASS %s\n", passwd);
 send(soc, buf, strlen(buf),0);
 recv_line(soc, buf, 512);
 if(strncmp(buf, "230", 3)){
  efree(&buf);
  return(1);
 }
 while(buf[3]=='-')recv_line(soc, buf, 512);
 efree(&buf);
 return(0);
}

int get_pasv_address(int soc, struct sockaddr_in * addr)
{
 char * buf = emalloc(255);
 char * t,*s;
 unsigned char * l;
 unsigned long  * a;
 int i;
 unsigned short * p;
 sprintf(buf, "PASV\n");
 send(soc, buf, strlen(buf), 0);
 bzero(buf, 255);
 bzero(addr, sizeof(struct sockaddr_in));
 recv_line(soc, buf, 255);
 if(strncmp(buf, "227", 3)){
   efree(&buf);
   return(1);
   }
 t = strchr(buf, '(');
 if(!t)return(1);
 t++;
 s = strchr(t, ',');
 if(!s)return(1);
 s[0] = 0;
 
 l = emalloc(6);
 l[0] = (unsigned char)atoi(t);
 s++;
 t = strchr(s, ',');if(!t)return(1);
 t[0]=0;
 l[1] = (unsigned char)atoi(s);
 t++;
 s = strchr(t, ',');if(!s)return(1);
 s[0]=0;
 l[2] = (unsigned char)atoi(t);
 s++;
 t = strchr(s, ',');if(!t)return(1);
 t[0]=0;
 l[3] = (unsigned char)atoi(s);
 t++;
 s = strchr(t, ',');if(!s)return(1);
 s[0]=0;
 l[4] = (unsigned char)atoi(t);
 s++;
 t = strchr(s, ')');if(!t)return(1);
 t[0]=0;
 l[5] = (unsigned char)atoi(s);
 a = (unsigned long*)l;
 p = (unsigned short*)(l+4);
 addr->sin_addr.s_addr = *a;
 addr->sin_port=*p;
 addr->sin_family = AF_INET;
 return(0);
}

int dangerous_password_file(char * passwd)
{
 char * safe_accounts[] = {"root", "toor", "daemon", "operator", "bin", 
 			   "games", "news", "man", "uucp", "xten", "pop",
                           "nobody", "ftp", "adm", "lp", "sync", "shutdown",
                           "halt", "mail", "postmaster"};
 
 int ok = 1;
 char * passwd_2 = emalloc(strlen(passwd)+1);
 
 strncpy(passwd_2, passwd, strlen(passwd));
 passwd = passwd_2;
 
 while(ok)
 {
  char * t = strchr(passwd, '\n');
  char * passwd_field;
  if(!t)break;
  t[0] = 0;
  passwd_field = strchr(passwd, ':');
  if(passwd_field)
  {
   int i;
   int j = 0;
   char * next;
   passwd_field[0]=0;
   passwd_field++;
   for(i=0;i<20;i++)if(!strcmp(passwd, safe_accounts[i]))j++;
   if(!j)ok=0;
   next = strchr(passwd_field, ':');
   if(next){
    next[0]=0;
    if(strlen(passwd_field)>1)ok = 0;
    }
  }
  else break;
  passwd = t+1;
 }
 return(ok?0:1);
}
  
 
